#!/usr/bin/env sh
# shellcheck source=onedatify/onedatify
. /dev/null

validate_msg_error=""

config_usage() {
cat <<EOF
onedatify config -p <admin password>

  -p, --password   admin password

EOF
}

config_main() {

  if [ $(( $# )) -eq 0 ]; then
    config_usage
  fi

  while [ $(( $# )) -ne 0 ] ; do
      case $1 in
          -h|-\?|--help)   # Call a "usage" function to display a synopsis, then exit.
              config_usage
              exit 0
              ;;
          show)
              shift
              config_load "$config_file_path"
              config_show "$@"
              return $?
              ;;
          validate)
              shift
              config_load "$config_file_path"
              config_validate "$@"
              return $?
              ;;
          edit)
              shift
              config_load "$config_file_path"
              config_edit "$config_file_path"
              return $?
              ;;
          interactive)
              shift
              config_interactive "$@"
              return $?
              ;;
          set)
              shift
              config_set "$@"
              return $?
              ;;
          -?*|*)
              printf 'WARN: Unknown option (ignored): %s\n' "$1" >&2
              exit 1
              ;;
      esac
      shift
  done
}

#######################################
# config set
# in the config file
#######################################
config_set_usage() {
cat <<EOF
TODO
EOF
}
config_set() {

  if [ $(( $# )) -eq 0 ]; then
    echo "No parameters found"
  fi

  config_set_template=0
  while [ $(( $# )) -ne 0 ] ; do
      case $1 in
          -h|-\?|--help)
              config_set_usage
              exit 0
              ;;
          -t|--template-config)
              #TODO: document this!
              config_set_template=1
              # TODO when empty argument
              # /usr/bin/onedatify: 97: shift: can't shift that many
              ;;
          -?*)
              printf 'storage_get WARN: Unknown option (ignored): %s\n' "$1" >&2
              exit 1
              ;;
          *)
              config_set_parameter="$1"
              ;;
      esac
      shift
  done

  if [ "$config_set_template" = 0 ]; then
    config_set_config_file="$config_file_path"
  else
    config_set_config_file="$template_config_file_path"
  fi

  # TODO: this cannot fail silently when there is no file!
  if [ -f "$config_set_config_file" ]; then
    config_set_parameters=$(echo "$config_set_parameter" | sed -e 's/\([^\/]\),/\1\n/g' )
    for f in $config_set_parameters; do
      config_set_parameters_key=${f%%=*}
      config_set_parameters_key_tmp=${f#$config_set_parameters_key}
      config_set_parameters_val=${config_set_parameters_key_tmp#=}
      eval "config_$config_set_parameters_key=$config_set_parameters_val"
      config_save "$config_set_parameters_key" "$config_set_config_file"
    done
  fi
}

#######################################
# config parameters expected to be found
# in the config file
#######################################
config_parameters="version name fqdn persistence_volume subdomain domain_name key cert cacert geo_latitude geo_longitude zone_fqdn admin_password \
              admin_email subdomain_delegation lets_encrypt_enabled onezone_registration_token"
#             onedatify_dir read_only storage_type storage_name import_enabled \
#             posix_mountPoint \
#             s3_hostname s3_bucketName s3_accessKey s3_secretKey s3_insecure \
#             ceph_username ceph_key ceph_monitorHostname ceph_clusterName ceph_poolName ceph_insecure\
#             gluster_hostname gluster_volume gluster_transport gluster_insecure"


#######################################
# config edit section
#
#######################################
config_edit() {
  config_edit_file="$1"
  config_edit_tmp_file=$(mktemp "${TMPDIR:-/tmp/}$(basename $0).XXXXXXXXXXXX")

  _config_edit_create_file() {
   cp "$config_edit_file" "$config_edit_tmp_file"

    sed -i -e "# Please edit the config below. Lines beginning with a '\\#' will be ignored, \\n\
# and an empty file will abort the edit. If an error occurs while saving, this file will be\\n\
# reopened with the relevant failures.\\n#" "$config_edit_tmp_file"

    error_line=5
    for var in $config_parameters; do
      eval tmp_val="\$config_$var"
      eval tmp_label="\$display_$var"
      if [ -n "$(type -t validate_${var})" ] && [ "$(type -t validate_${var})" = function ]; then
        if validate_$var "$tmp_val" ; then
          error=""
        else
          error="ERROR: $var $validate_msg_error"
        fi
      else 
        if validate_empty "$tmp_val" ; then
          error=""
        else
          error="ERROR: $var $validate_msg_error"
        fi
      fi
      if [ "$error" != "" ]; then
        sanitized_error=$(echo "$error" | sed -e 's/[\/&]/\\&/g')
        sed -i "${error_line}i # $sanitized_error" "$config_edit_tmp_file"
        error_line=$((error_line+1))
      fi
      sanitized_var=$(echo "$var" | sed -e 's/[\/&]/\\&/g')
      sanitized_val=$(echo "$tmp_val" | sed -e 's/[\/&]/\\&/g')
      sed -i -e "s/^\([[:space:]]*${sanitized_var}[[:space:]]*=\).*/\1$sanitized_val/" "$config_edit_tmp_file"
    done
    sed -i -e "${error_line}i #" "$config_edit_tmp_file"
  }


  while : ; do 
    _config_edit_create_file
    vim "$config_edit_tmp_file"
    config_load "$config_edit_tmp_file"
    if config_validate 1 ; then
      config_save "$config_parameters" "$config_edit_file"
      rm "$config_edit_tmp_file"
      return 0
    else
      while : ; do
        config_edit_continue_default="y"
        _message 'The configuration contains errors.\n'
        _message "Do you want to continue editing (y), abort (n), or save erroneous configuration anyway (s) (default: $config_edit_continue_default)?: "
        read -r config_edit_continue </dev/tty
        [ "$config_edit_continue" = "" ] && config_edit_continue=$config_edit_continue_default
        if [ "$config_edit_continue" = "y" ] ; then
          config_load "$config_edit_tmp_file"
          break
        elif [ "$config_edit_continue" = "s" ]; then
          config_save "$config_parameters" "$config_edit_file"
          echo "Your changes were saved to $config_edit_file."
          echo "WARNING: Your configuration still contains errors! Oneprovider will not start!"
          return 1
        elif [ "$config_edit_continue" = "n" ]; then
          echo "A copy of your changes has been stored to $config_edit_tmp_file"
          echo "error: Edit cancelled, no valid changes were saved."
          return 1
        else
          echo "no valid option <$config_edit_continue>"
          continue
        fi
      done
    fi
  done
}

#######################################
# config validate section
#
#######################################
config_validate() {
  quiet="$1"
  [ "$quiet" = "" ] && quiet=0

  errors=""
  for var in $config_parameters; do
    eval tmp_val="\$config_$var"
    eval tmp_label="\$display_$var"
    if [ -n "$(type -t validate_${var})" ] && [ "$(type -t validate_${var})" = function ]; then
      if validate_$var "$tmp_val" ; then
        error=""
      else
        errors="$errors\\nERROR: $var $validate_msg_error"
      fi
    else 
      if validate_empty "$tmp_val" ; then
        error=""
      else
        errors="$errors\\nERROR: $var $validate_msg_error"
      fi
    fi
  done
  # if [ "$errors" != "" ]; then
  #   if [ $quiet -eq 0 ]; then
  #     echo "Found errors in the configuration, please fix them to proceed:"
  #     echo "$errors"
  #   fi
  #   return 1
  # fi
  return 0
}

#######################################
# config show section
#
#######################################
# shellcheck disable=SC2034
config_show() {
  display_version="version:"
  display_name="Oneprovider name:"
  display_fqdn="Oneprovider Fully qualified domain name (FQDN):"
  display_persistence_volume="persistence_volume:"
  display_hostname="hostname:"
  display_domain_name="domain_name:"
  display_key="key:"
  display_cert="cert:"
  display_cacert="cacert:"
  display_geo_latitude="geo_latitude:"
  display_geo_longitude="geo_longitude:"
  display_zone_fqdn="zone_fqdn:"
  display_onezone_registration_token="onezone_registration_token:"
  display_onedatify_dir="onedatify_dir:"
  display_read_only="read_only:"
  display_admin_password="admin_password:"
  display_admin_email="admin_email:"
  display_storage_type="storage_type:"
  display_storage_name="storage_name:"
  display_import_enabled="import_enabled:"
  display_posix_mountPoint="posix_mountPoint:"
  display_s3_hostname="s3_hostname:"
  display_s3_bucketName="s3_bucketName:"
  display_s3_accessKey="s3_accessKey:"
  display_s3_secretKey="s3_secretKey:"
  display_s3_insecure="s3_insecure:"
  display_ceph_username="ceph_username:"
  display_ceph_key="ceph_key:"
  display_ceph_monitorHostname="ceph_monitorHostname:"
  display_ceph_clusterName="ceph_clusterName:"
  display_ceph_poolName="ceph_poolName:"
  display_ceph_insecure="ceph_insecure:"
  display_gluster_hostname="gluster_hostname:"
  display_gluster_volume="gluster_volume:"
  display_gluster_transport="gluster_transport:"
  display_gluster_insecure="gluster_insecure:"

  for var in $config_parameters; do
    eval tmp_val="\$config_$var"
    eval tmp_label="\$display_$var"

     if [ -n "$(type -t validate_${var})" ] && [ "$(type -t validate_${var})" = function ]; then
      if validate_$var "$tmp_val" ; then
        error=""
      else
        error="[ERROR: $validate_msg_error]"
      fi
    else 
      if validate_empty "$tmp_val" ; then
        error=""
      else
        error="[ERROR: $validate_msg_error]"
      fi
    fi

    if [ "$error" = "" ]; then
      text_color=$(tput setaf 2)
    else
      text_color=$(tput setaf 1)
    fi
    normal=$(tput sgr0)

    $sh_c "printf '%-20s %-80s %-30s\\n' \"$tmp_label\" \"${text_color}$tmp_val${normal}\" \"$error\"" 
  done
}

#######################################
# config interactive section
#
#######################################

config_interactive() {

  config_interactive_onezone_fqdn=1
  config_interactive_name=1
  config_interactive_password=1
  config_interactive_admin_email=1
  config_interactive_persistence_volume=1
  config_interactive_coordinates=1
  config_interactive_subdomain_delegation=1
  config_interactive_subdomain=1
  config_interactive_lets_encrypt_enabled=1
  config_interactive_fqdn=1
  config_interactive_certs=1

  config_interactive_exclude=""
  while [ $(( $# )) -ne 0 ] ; do
      case $1 in
          -h|-\?|--help)
              storage_usage_get
              exit 0
              ;;
          -e|--exclude)
              config_interactive_exclude="$2"
              shift
              # TODO when empty argument
              # /usr/bin/onedatify: 97: shift: can't shift that many
              ;;
          -?*|*)
              printf 'storage_get WARN: Unknown option (ignored): %s\n' "$1" >&2
              exit 1
              ;;
      esac
      shift
  done

  config_interactive_excludes=$(echo "$config_interactive_exclude" | sed -e 's/\([^\/]\),/\1\n/g' )
  for f in $config_interactive_excludes; do
    # TODO: MAKE IT SAFE!
    eval "config_interactive_$f=0"
  done

  if [ -f "$config_file_path" ] ; then
    message ''
    message '########################################'
    message "An old configuration has been detected:"

    config_load "$config_file_path"
    echo ""
    $sh_c "cat $config_file_path"
    echo ""
    message "########################################"
    message ""
    clean=""
    while [ "$clean" != "n" ] && [ "$clean" != "y" ] ; do
      message_imp "An existing Oneprovider configuration has been detected on your machine (see above)."
      message_imp "WARNING: If you proceed, the existing installation will be completely deleted and a clean one will be performed."
      _message "Do you want to continue (y/n)?: "
      read -r clean </dev/tty
    done
    if [ "$clean" = "y" ]; then
        message "Gracefully deregistering Oneprovider from Onezone ${config_zone_fqdn}..."
        if ! curl -k --fail --silent -u "onepanel:${config_admin_password}" -sS --tlsv1.2 -X DELETE "https://localhost:9443/api/v3/onepanel/provider" 2>/dev/null; then
          if [ -f "$config_persistence_volume/etc/op_worker/provider_root_macaroon.txt" ] ; then
            oneprovider_root_macaroon_file="$config_persistence_volume/etc/op_worker/provider_root_macaroon.txt"
            oneprovider_id=$(cat "$oneprovider_root_macaroon_file" | tr -d '\n' | sed  's/.*provider_id\s*=>\s*<<"\([[:alnum:]]*\)">>.*/\1/')
            deregistration_token=$(cat "$oneprovider_root_macaroon_file" | tr -d '\n' | sed  's/.*root_macaroon\s*=>\s*<<"\([[:alnum:]]*\)">>.*/\1/')
            if ! curl  -k --fail --silent -H "macaroon:$deregistration_token" -sS --tlsv1.2 -X DELETE "https://${config_zone_fqdn}/api/v3/onezone/providers/$oneprovider_id"; then
              message_imp "WARNING: Deregistration failed. Your Oneprovider is probably already deregistered or missing its persistence files."
            fi
          else
            message_imp "WARNING: Deregistration failed. Your Oneprovider is probably already deregistered or missing its persistence files."
          fi
        fi

        message "Stopping Oneprovider docker container..."
        $sh_c "systemctl -q is-active onedatify.service >/dev/null && systemctl stop onedatify.service"
        message "Removing Oneprovider docker container..."
        $sh_c "docker rm -f onedatify-oneprovider-1" 
        message "Removing Oneprovider configuration directory..."
        $sh_c "rm -rf $config_persistence_volume"
        $sh_c "rm $onedatify_docker_compose_path"
        $sh_c "cp $template_config_file_path $config_file_path"
    else
        message "This script can only perform installation on a clean system."
        exit 1
    fi
  else
    $sh_c "cp $template_config_file_path $config_file_path"
    config_load "$config_file_path"
  fi

  message ''
  message '########################################'
  message 'You will now be asked to provide configuration parameters:'
  message ''

  # Those functions depend on each other
  # so care when changing order
  message ''
  message '########## Basic parameters ###########'
  [ "$config_interactive_onezone_fqdn" = "1" ] && get_onezone_fqdn
  [ "$config_interactive_name" = "1" ] && get_name
  [ "$config_interactive_password" = "1" ] && get_password
  [ "$config_interactive_admin_email" = "1" ] && get_admin_email
  [ "$config_interactive_persistence_volume" = "1" ] && get_persistence_volume
  [ "$config_interactive_coordinates" = "1" ] && get_coordinates

  message ''
  message '########### Domain and certificates #############'
  message "You may provide your own domain for the Oneprovider, or choose to acquire a FQDN using Onezone's subdomain delegation."
  message "In the latter case, your Oneprovider will be assigned a subdomain in the Onezone's domain, like the following:"
  message "    <subdomain>.${config_zone_fqdn}"
  message "and a Let's Encrypt certificate will be automatically generated for you."
  message ""
  # TODO: add automatic detection if onezone supports subdomain delegation
  [ "$config_interactive_subdomain_delegation" = "1" ] && get_subdomain_delegation
  if [ "$config_subdomain_delegation" = "true" ]; then
    [ "$config_interactive_subdomain" = "1" ] && get_subdomain
    
  else
    [ "$config_interactive_fqdn" = "1" ] && get_fqdn
    [ "$config_interactive_lets_encrypt_enabled" = "1" ] && get_lets_encrypt_enabled
    if [ "$config_lets_encrypt_enabled" = "false" ]; then
      [ "$config_interactive_certs" = "1" ] && get_certs
    

    local_copy_of_docker_compose_yaml=$(echo "$local_copy_of_docker_compose_yaml" | sed  -e "s%volumes:%volumes:\\n\
    # Line 17 certs\\n\
       # Oneprovider certificate key\\n\
       - \"\$key:/etc/op_panel/certs/key.pem:ro\"\\n\
       # Oneprovider public certificate\\n\
       - \"\$cert:/etc/op_panel/certs/cert.pem:ro\"\\n\
       # Certificate of public certificate signing authority\\n\
       - \"\$cacert:/etc/op_panel/cacerts/cacert.pem:ro\"\\n\
       # Certificate of public certificate signing authority (same as above)\\n\
       - \"\$cacert:/etc/op_worker/cacerts/cacert.pem:ro\"\\n\
    # Line 18 certs\\n\
       # Oneprovider certificate key\\n\
       - \"\$key:/etc/op_panel/certs/web_key.pem:ro\"\\n\
       # Oneprovider public certificate\\n\
       - \"\$cert:/etc/op_panel/certs/web_cert.pem:ro\"\\n\
       # Certificate of public certificate signing authority\\n\
       - \"\$cacert:/etc/op_panel/certs/web_chain.pem:ro\"\\n\
       # Additional, trusted CA certificates (any filename is accepted)\\n\
       #- \"\${extra_cacerts}:/etc/op_panel/cacerts/customCaBundle.pem\"\\n\
       %")
    fi
  fi

  config_save "$config_parameters" "$config_file_path"
  $sh_c "tee $onedatify_docker_compose_path > /dev/null" << EEOF
$local_copy_of_docker_compose_yaml
EEOF
}

#######################################
# config save and load
#
#######################################

config_load() {
  #cat $1

  # problem: https://stackoverflow.com/questions/12916352/shell-script-read-missing-last-line
  # hack: https://stackoverflow.com/a/34945087
  config_load_last_line=$(tail -c 1 "$1")
  [ "$config_load_last_line" != "" ] && sed -i -e '$a\' "$1"
  while read -r field || [ -n "$field" ] ; do
    [ "$field" = "" ] && continue ;
    if [ $(expr "$field" : "[[:space:]]*#.*") -ne 0 ] ; then continue ; fi
    config_load_tmp=$(echo "$field" | awk 'BEGIN { FS = "=" }; { print $1 }')
    key="config_${config_load_tmp}"
    value="${field#*=}"
    export "$key=$value"
  done < "$1"
 
}

config_save() {
  config_save_parameters="$1"
  config_save_file="$2"
  for var in $config_save_parameters; do
    eval tmp_val="\$config_$var"
    eval tmp_label="\$display_$var"
    sanitized_var=$(echo "$var" | sed -e 's/[\/&]/\\&/g')
    sanitized_val=$(echo "$tmp_val" | sed -e 's/[\/&]/\\&/g')
    $sh_c "sed -i -e \"s/^\([[:space:]]*${sanitized_var}[[:space:]]*=\).*/\1$sanitized_val/\" \"$config_save_file\""
  done
}

#######################################
# Generic config validators
#
#######################################

validate_empty() {
  validate_msg_error="empty field"
  [ "$1" != "" ] || [ "$(${_expr} "$1" : "^[[:space:]]*$")" -ne "0" ]
}

validate_boolean() {
  validate_msg_error="bad value, has to be true or false"
  [ "$1" != "true" ] || [ "$1" != "false" ]
}

validate_coordinate() {
  validate_msg_error="bad format"
  [ "${#1}" = "$(${_expr} "$1" : '-\?[0-9]\+\|-\?[0-9]\+\.[0-9]\+')" ]
}

validate_domain() {
  validate_msg_error="bad format: [a-z0-9]([-a-z0-9]*[a-z0-9])?"
  validate_domain_tmp=$(${_expr} "$1" : '\([a-z0-9]\([-a-z0-9]*[a-z0-9]\)\?\)')
  [ "$validate_domain_tmp" != "" ] && [ "$1" = "$validate_domain_tmp"  ]
}

validate_file_exists() {
  # File exists and it's size is greater then zero
  validate_msg_error="file does not exists"
  [ -f "$1" ] || return 1
  validate_msg_error="file is empty"
  [ -s "$1" ] || return 1
}


#######################################
# Config setters with their validators
#
#######################################

get_onezone_fqdn() {
  ######
  # Ask for onezone FQDN or ip
  ######
  if [ "$config_onezone_fqdn" = "" ] ; then
    config_onezone_fqdn_default="onezone.onedata.org"
  else
    config_onezone_fqdn_default="$config_onezone_fqdn"
  fi
  _message "The fully qualified domain name (FQDN) or ip address of the Onezone where the Oneprovider will be registered (default: $config_onezone_fqdn_default): "
  read -r config_onezone_fqdn </dev/tty
  [  "$config_onezone_fqdn" = "" ] && config_onezone_fqdn="$config_onezone_fqdn_default"
}
validate_onezone_fqdn() {
  validate_empty "$1"
}

get_name() {
  #####
  # Ask for Oneprovider display name
  ##### 
  if [ "$config_name" = "" ] ; then
    config_name_default=$($sh_c "hostname")
  else
    config_name_default=$config_name
  fi
  _message "Enter the display name for your Oneprovider (default: $config_name_default): "
  read -r config_name </dev/tty
  [  "$config_name" = "" ] && config_name="$config_name_default"
}
validate_name() {
  validate_empty "$1"
}

get_persistence_volume() {
  ######
  # Ask for a directory to store provider config and metadata
  ######
  if [ "$config_persistence_volume" = "" ] ; then
    config_persistence_volume_default="$onedatify_config_dir/oneprovider_conf"
  else
    config_persistence_volume_default="$config_persistence_volume"
  fi
  _message "Absolute path where to keep Oneprovider configuration and metadata (default: $config_persistence_volume_default): "
  read -r config_persistence_volume </dev/tty
  [  "$config_persistence_volume" = "" ] && config_persistence_volume="$config_persistence_volume_default"
}
validate_persistence_volume() {
  validate_empty "$1"
}

get_coordinates() {
  ######
  # Ask for coordinates
  ######
  config_coordinates=$($sh_c "$curl  http://api.ipstack.com/check?access_key=45e12e9956b86e77b32930cd223f1ef4 | tr ',' '\n' | grep -E 'longitude|latitude' | cut -d ':' -f 2" | tr -d ' ')

  if [ "$config_coordinates" != "" ]; then
    [ "$config_geo_latitude" = "" ] && config_geo_latitude=$(echo "$config_coordinates" | sed -n 1p)
    [ "$config_geo_longitude" = "" ] && config_geo_longitude=$(echo "$config_coordinates" | sed -n 2p)
  else
    message_imp "Automatic detection of latitude and longitude failed. Cannot connect to http://api.ipstack.com."
  fi
  
  config_geo_latitude_default=$config_geo_latitude
  while true ; do
    _message "Enter the latitude of your Oneprovider (auto-detected default: $config_geo_latitude_default): "
    read -r config_geo_latitude </dev/tty
    if [ "$config_geo_latitude" =  "" ]; then
      config_geo_latitude=$config_geo_latitude_default
    fi
    if validate_coordinate "$config_geo_latitude" ; then
      break 
    else 
      message_imp "The format of the coordinates ('$config_geo_latitude') needs to mach the regular expression '[-0-9]*[0-9][0-9]*.[0-9][0-9]*'. Please try again."
    fi
  done

  config_geo_longitude_default=$config_geo_longitude
  while true ; do
    _message "Enter the longitude of your Oneprovider (auto-detected default: $config_geo_longitude_default): "
    read -r config_geo_longitude </dev/tty
    if [ "$config_geo_longitude" =  "" ]; then
      config_geo_longitude=$config_geo_longitude_default
    fi
    if validate_coordinate "$config_geo_longitude" ; then
      break 
    else 
      message_imp "The format of the coordinates ('$config_geo_longitude') needs to mach the regular expression '[-0-9]*[0-9][0-9]*.[0-9][0-9]*'. Please try again."
    fi
  done
}

validate_geo_latitude() {
  validate_coordinate "$1"
}
validate_geo_longitude() {
  validate_coordinate "$1"
}
get_fqdn() {
  ######
  # Ask for provider FQDN
  ######
  #config_fqdn=$($sh_c "$curl http://ipinfo.io/ip")
  
  if [ "$config_fqdn" = "" ] ; then
    config_fqdn_default="$($sh_c "hostname -f")"
  else
    config_fqdn_default=$config_fqdn
  fi
  _message "Enter the fully qualified domain name (FQDN) of your Oneprovider (auto-detected default: $config_fqdn_default): "
  read -r config_fqdn </dev/tty
  [  "$config_fqdn" = "" ] && config_fqdn="$config_fqdn_default"

  config_subdomain=$(echo "$config_fqdn" | cut -d '.' -f 1)
  config_domain_name=$(echo "$config_fqdn" | cut -d '.' -f 2-)
  config_check_and_fix_etc_hosts "$config_subdomain" "$config_domain_name"
}
validate_fqdn() {
  validate_empty "$1"
}
validate_hostname() {
  validate_empty "$1"
}
validate_domain_name() {
  validate_empty "$1"
}

get_password() {
  ######
  # Ask for or generate a password
  ######
  if [ "$config_admin_password" = "" ] ; then
    config_admin_password_default=$(head /dev/urandom | tr -dc A-Za-z0-9 | head -c 13)
  else
    config_admin_password_default=$config_admin_password
  fi
  _message "Enter a new emergency passphrase for the Oneprovider admin panel or leave it blank to be autogenerated: "
  read -r config_admin_password </dev/tty
  [  "$config_admin_password" = "" ] && config_admin_password="$config_admin_password_default"
  message "The emergency passphrase to log in to Oneprovider admin panel is: $config_admin_password"
}
validate_password() {
  validate_empty "$1"
}

get_admin_email() {
  ######
  # Ask for or generate a password
  ######
  config_admin_email_default=$config_admin_email
  config_admin_email=""
  while true ; do
    _message "Enter the email address that will be used as emergency contact for this Oneprovider (default: $config_admin_email_default): "
    read -r config_admin_email </dev/tty
    [ "$config_admin_email" = "" ] && config_admin_email=$config_admin_email_default
    if [ "$config_admin_email" = "" ] || ! validate_admin_email "$config_admin_email" ; then
        message_imp "$validate_msg_error"
    else
      break
    fi
  done
}
validate_admin_email() {
  regex="\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,6}\b"
  validate_msg_error="Invalid email format. The email must match: [a-zA-Z0-9]+(.[a-zA-Z0-9]+)*@[a-zA-Z0-9]+(.[a-zA-Z0-9]+)+ "
  #[ "$1" = "$(${_expr} "$1" : '\([a-zA-Z0-9]\+\(\.[a-zA-Z0-9]\+\)*@[a-zA-Z0-9]\+\(\.[a-zA-Z0-9]\+\)\+\)')" ]
  echo "$1" | grep -E -s "$regex" > /dev/null
}

get_subdomain_delegation() {
  if [ "$config_subdomain_delegation" = "true" ]; then
    config_subdomain_delegation_default="y"
  elif [ "$config_subdomain_delegation" = "false" ]; then
    config_subdomain_delegation_default="n"
  else 
    config_subdomain_delegation_default="y"
  fi
  while true ; do
    _message "Do you want to use subdomain delegation to acquire a FQDN for your Oneprovider? (y/n, default: $config_subdomain_delegation_default)?: "
    read -r config_subdomain_delegation </dev/tty
    [ "$config_subdomain_delegation" =  "" ] && config_subdomain_delegation=$config_subdomain_delegation_default
    if [ "$config_subdomain_delegation" != "n" ] || [ "$config_subdomain_delegation" != "y" ]; then
      break
    fi
  done
  if [ "$config_subdomain_delegation" = "y" ]; then
      config_subdomain_delegation=true
  else
      config_subdomain_delegation=false
  fi
}
validate_subdomain_delegation() {
  validate_empty "$1" && validate_boolean "$1"
}
get_subdomain() {
  ######
  # Ask for subdomain
  ######
  #config_subdomain=$($sh_c "$curl http://ipinfo.io/ip")

  if [ "$config_subdomain" = "" ] ; then
    config_subdomain_default=$($sh_c "hostname")
  else
    config_subdomain_default=$config_subdomain
  fi
  while true ; do
    _message "Enter the subdomain for your Oneprovider (auto-detected default: $config_subdomain_default): "
    read -r config_subdomain </dev/tty
    [ "$config_subdomain" =  "" ] && config_subdomain=$config_subdomain_default
    if ! validate_subdomain "$config_subdomain"; then
      message_imp "The format of the subdomain must mach the regular expression '[a-z0-9]([-a-z0-9]*[a-z0-9])?'. Please try again."
    elif ! validate_subdomain_availability "${config_subdomain}.${config_zone_fqdn}" "$config_zone_fqdn" ; then
      message_imp "The subdomain \"$config_subdomain\" is already registered in Onezone $config_zone_fqdn. Please choose a different subdomain." 
    else
      break
    fi
  done

  [  "$config_subdomain" = "" ] && config_subdomain="$config_subdomain_default"

  config_domain_name="$config_zone_fqdn"
  # config_check_and_fix_etc_hosts "$config_subdomain" "$config_domain_name"
  config_fqdn="${config_subdomain}.${config_domain_name}"

  _message "The FQDN of your Oneprovider is: "; message "$config_fqdn" 
}
validate_subdomain() {
  validate_domain "$1" && validate_empty "$1"
}
validate_subdomain_availability() {
  ! host -r "$1" "$2" >/dev/null 2>&1
}
validate_hostname() {
  validate_empty "$1"
}
validate_domain_name() {
  validate_empty "$1"
}

get_lets_encrypt_enabled() {
  if [ "$config_lets_encrypt_enabled" = "true" ]; then
    config_lets_encrypt_enabled_default="y"
  elif [ "$config_lets_encrypt_enabled" = "false" ]; then
    config_lets_encrypt_enabled_default="n"
  else
    config_lets_encrypt_enabled_default="n"
  fi
  while true ; do
    _message "Do you want to obtain a Let's Encrypt certificate for your Oneprovider automatically (y/n, default: $config_lets_encrypt_enabled_default)? "
    read -r config_lets_encrypt_enabled </dev/tty
    [ "$config_lets_encrypt_enabled" =  "" ] && config_lets_encrypt_enabled=$config_lets_encrypt_enabled_default
    if [ "$config_lets_encrypt_enabled" = "n" ] || [ "$config_lets_encrypt_enabled" = "y" ]; then
      break
    fi
  done
  if [ "$config_lets_encrypt_enabled" = "y" ]; then
      config_lets_encrypt_enabled=true
      message_imp "INFO: Let's Encrypt allows to create only 5 duplicate certificates per week! Use this option with care!"
  else
      config_lets_encrypt_enabled=false
  fi
}
validate_lets_encrypt_enabled() {
  validate_empty "$1" && validate_boolean "$1"
}

get_certs() {

  ######
  # Ask for certs
  ######
  letsencrypt_prefix="/etc/letsencrypt/live/$config_fqdn"

  config_key_default=$config_key
  [ "$config_key_default" = "" ] && [ -f "$letsencrypt_prefix/privkey.pem" ] && config_key_default="$letsencrypt_prefix/privkey.pem"
  config_key=""
  while true ; do
    _message "Enter the path to the private key for domain $config_fqdn (default: $config_key_default): "
    read -r config_key </dev/tty
    [ "$config_key" = "" ] && config_key=$config_key_default
    if [ -f "$config_key" ] ; then
      break
    else
      message_imp "There is no file under $config_key the path you provided, please try again."
    fi
  done

  config_cert_default=$config_cert
  [ "$config_cert_default" = "" ] && [ -f "$letsencrypt_prefix/cert.pem" ] && config_cert_default="$letsencrypt_prefix/cert.pem"
  config_cert=""
  while true ; do
    _message "Enter the path to the public certificate for domain $config_fqdn (default: $config_cert_default): "
    read -r config_cert </dev/tty
    [ "$config_cert" = "" ] && config_cert=$config_cert_default
    if [ -f "$config_cert" ] ; then
      break
    else
      message_imp "There is no file at the path you provided, please try again."
    fi
  done

  config_cacert_default=$config_cacert
  [ "$config_cacert_default" = "" ] && [ -f "$letsencrypt_prefix/chain.pem" ] && config_cacert_default="$letsencrypt_prefix/chain.pem"
  config_cacert=""
  while true ; do
    _message "Enter the path to the certificate chain of the signing authority of your certificate (default: $config_cacert_default): "
    read -r config_cacert </dev/tty
    [ "$config_cacert" = "" ] && config_cacert=$config_cacert_default
    if [ -f "$config_cacert" ] ; then
      break
    else
      message_imp "There is no file at the path you provided, please try again."
    fi
  done
}
validate_key() {
  validate_empty "$1" && validate_file_exists "$1"
}
validate_cert() {
  validate_empty "$1" && validate_file_exists "$1"
}
validate_cacert() {
  validate_empty "$1" && validate_file_exists "$1"
}

# TODO: validate is it a proper docker image tag
validate_version() {
  validate_empty "$1"
}


config_check_and_fix_etc_hosts() {
  config_check_and_fix_etc_hosts_hostname="$1"
  config_check_and_fix_etc_hosts_domain="$2"
  if [ "$(hostname)" != "$config_check_and_fix_etc_hosts_hostname" ] || [ "$(hostname -f)" != "$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain" ]; then
    message_imp "ERROR: The configuration of your VM is incorrect:"
    message_imp ''
    message_imp "'hostname' should return '$config_check_and_fix_etc_hosts_hostname'"
    message_imp "'hostname -f' should return '$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain'"
    message_imp ''
    message_imp 'The needed changes are:'
    message_imp '1) set the proper hostname of the machine to match the subdomain'
    message_imp "   use command: 'hostnamectl set-hostname $config_check_and_fix_etc_hosts_hostname'"
    message_imp '2) replace the current content of /etc/hosts:'
    message_imp ''
    $sh_c "cat /etc/hosts"
    message_imp ''

    # Read the content of /etc/hosts
    config_fqdn_line_etc_hosts="$(cat /etc/hosts)"

    # First lets look for not commented out line with 127.0.1.1
    config_check_and_fix_etc_hosts_line_number=$(grep -n "^127.0.1.1" /etc/hosts |cut -f1 -d:)
    config_fqdn_line_comment_out=''
    if [ "$config_check_and_fix_etc_hosts_line_number" != "" ]; then
      # if it's there we need to comment it out
      config_fqdn_line_etc_hosts=$(echo "$config_fqdn_line_etc_hosts" | sed "${config_check_and_fix_etc_hosts_line_number}s/.*/#\0/")

      # there is also a chance that a line we need is already in /etc/hosts, just commented out
      config_check_and_fix_etc_hosts_line_number_exists=$(grep -n "^[[:space:]]*#[[:space:]]*127.0.1.1[[:space:]]*$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain[[:space:]]*$config_check_and_fix_etc_hosts_hostname" /etc/hosts |cut -f1 -d:)
      if [ "$config_check_and_fix_etc_hosts_line_number_exists" != "" ]; then
        # if it is there, we need to comment otu bad line and uncomment this one
        config_fqdn_line_etc_hosts=$(echo "$config_fqdn_line_etc_hosts" |sed "s/^[[:space:]]*#[[:space:]]*\(127.0.1.1[[:space:]]*$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain[[:space:]]*$config_check_and_fix_etc_hosts_hostname\)/\1/")
      else
        # if ot was not there lets jsut add new line after the line we commented out
        config_fqdn_line_etc_hosts=$(echo "$config_fqdn_line_etc_hosts" | sed "${config_check_and_fix_etc_hosts_line_number}s/.*/${config_fqdn_line_comment_out}\0\n127.0.1.1\t$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain\t$config_check_and_fix_etc_hosts_hostname/")
      fi
    else
      # if it's not here, lets look for localhost and add our line after it
      config_check_and_fix_etc_hosts_line_number=$(grep -n "^127.0.0.1" /etc/hosts |cut -f1 -d:)
      if [ "$config_check_and_fix_etc_hosts_line_number" != "" ]; then
        config_fqdn_line_etc_hosts=$(echo "$config_fqdn_line_etc_hosts" | sed "${config_check_and_fix_etc_hosts_line_number}s/.*/\0\n127.0.1.1\t$config_check_and_fix_etc_hosts_hostname.$config_check_and_fix_etc_hosts_domain\t$config_check_and_fix_etc_hosts_hostname/")
      else
        message_imp "ERROR: Your /etc/hosts seems to contain errors."
      fi
    fi

    message_imp "with the one below (changed lines are in color): "
    message_imp ''
    echo "$config_fqdn_line_etc_hosts" | grep --color -E  "^127.0.1.1.*|$"
    message_imp ''

    while true ; do
      _message "Do you want to apply above changes automatically? (y/n, default: y)? "
      read -r config_fqdn_apply_changes </dev/tty
      [ "$config_fqdn_apply_changes" =  "" ] && config_fqdn_apply_changes=y
      if [ "$config_fqdn_apply_changes" = "n" ] || [ "$config_fqdn_apply_changes" = "y" ]; then
        break
      fi
    done
    if [ "$config_fqdn_apply_changes" = "y" ]; then
      $sh_c "hostnamectl set-hostname $config_check_and_fix_etc_hosts_hostname"
      $sh_c "tee /etc/hosts > /dev/null" << EEOF
$config_fqdn_line_etc_hosts
EEOF
    else
      message_imp "ERROR: Cannot continue with incorrect VM configuration."
      exit 1
    fi

  fi
}

validate_disk_size() {
  validate_msg_error="Invalid format. The value must match: [1-9][0-9]*|[1-9][0-9]*B|[1-9][0-9]*[KMGTPE]iB|[1-9][0-9]*.[0-9][0-9]*[KMGTPE]iB"
  [ "${#1}" = "$(${_expr} "$1" : '[1-9][0-9]*\|[1-9][0-9]*B\|[1-9][0-9]*[KMGTPE]iB\|[1-9][0-9]*.[0-9][0-9]*[KMGTPE]iB')" ]
}

convert_disk_size() {
    convert_disk_size_value=$1
    case $convert_disk_size_value in
        *KiB)
            power=1
            ;;
        *MiB)
            power=2
            ;;
        *GiB)
            power=3
            ;;
        *TiB)
            power=4
            ;;
        *PiB)
            power=5
            ;;
        *EiB)
            power=6
            ;;
        *B)
            power=0
            convert_disk_size_value="${convert_disk_size_value}BB"
            ;;
        *)
            power=0
            convert_disk_size_value="${convert_disk_size_value}BBB"
    esac
    echo "${convert_disk_size_value%???}" "$power" | awk '{printf "%4.0f\n",$1*1024^$2}'
}

convert_back_disk_size() {
    convert_back_disk_size_value=$1
    #if [ "$convert_back_disk_size_value" -ge 1180591620717411303424 ]; then
    #  power=1180591620717411303424; unit="EiB"
    if [ "$convert_back_disk_size_value" -ge 1152921504606846976 ]; then
      power=1152921504606846976; unit="EiB"
    elif [ "$convert_back_disk_size_value" -ge 1125899906842624 ]; then
      power=1125899906842624; unit="PiB"
    elif [ "$convert_back_disk_size_value" -ge 1099511627776 ]; then
      power=1099511627776; unit="TiB"
    elif [ "$convert_back_disk_size_value" -ge 1073741824 ]; then
      power=1073741824; unit="GiB"
    elif [ "$convert_back_disk_size_value" -ge 1048576 ]; then
      power=1048576; unit="MiB"
    elif [ "$convert_back_disk_size_value" -ge 1024 ]; then
      power=1024; unit="KiB"
    else 
      power=1; unit="B"
    fi
    converted_back_disk_size_value=$(echo "$convert_back_disk_size_value" "$power" | awk '{printf "%4.4f\n",$1/$2}')
    echo "${converted_back_disk_size_value}${unit}"
}