#!/usr/bin/env sh
# shellcheck source=onedatify/onedatify

logs_usage() {
cat <<EOF
Show logs and copy logs from running instance of the Oneprovider.

${0##*/} logs [ show | copy ] [ --help ]

  show        tail and folow logs of Oneprovider 
  cp, copy    copy all the oneprovider logs to oneprovider_logs.tgz located in your local directory

EOF
}

logs_main() {

  if [ $(( $# )) -eq 0 ]; then
    logs_show
  fi

  while [ $(( $# )) -ne 0 ] ; do
      case $1 in
          -h|-\?|--help)   # Call a "usage" function to display a synopsis, then exit.
              logs_usage
              exit 0
              ;;
          show)
              shift
              logs_show "$@"
              exit $?
              ;;
          cp|copy)
              shift
              logs_copy "$@"
              exit $?
              ;;
          -?*|*)
              printf 'WARN: Unknown option (ignored): %s\n' "$1" >&2
              exit 1
              ;;
      esac
      shift
  done


  journalctl -u onedatify.service -f -n 1000
}

#######################################
# logs show 
#
#######################################
logs_show() {
  journalctl -u onedatify.service -f -n 1000
}

#######################################
# logs copy 
#
#######################################
logs_copy() {
  logs_copy_script=$(cat <<EEOF
#!/usr/bin/env bash
rm /oneprovider_logs.tgz
tar cvzfh /oneprovider_logs.tgz -C /var/log op_panel op_worker
EEOF
  )
  echo "$logs_copy_script" | docker exec \
  -i \
  onedatify-oneprovider-1\
  bash -c \
  'echo "-k" > ~/.curlrc ; cat > script.sh ; chmod +x script.sh ; ./script.sh'

  docker cp onedatify-oneprovider-1:oneprovider_logs.tgz oneprovider_logs.tgz
}
